import json
import boto3
import time
from botocore.exceptions import ClientError
from broly.column import Col, DateTime
from copy import deepcopy
from pypika import MySQLQuery as Query, Table, Order

client = boto3.client('rds-data', region_name='us-east-1')

class Model:

    created_at = DateTime(default_value='CURRENT_TIMESTAMP(6)')
    updated_at = DateTime(default_value='CURRENT_TIMESTAMP(6)', on_update='CURRENT_TIMESTAMP(6)')

    def __init__(self, *args, **kwargs):
        self.database = self.__get_database_name()
        self.secret_arn = self.__get_secret_arn()
        self.resource_arn = self.__get_resource_arn()
        self.table_name = self.__get_table_name()
        self.columns = self.__filter_columns()
        self.primary_key = [key for key in self.get_columns() if getattr(self,key).is_primary_key()][0]
        self.instance_cols = {}
        for col in self.columns:
            self.instance_cols[col] = deepcopy(getattr(self,col))
            value = getattr(self, col).default_value
            if col in kwargs:
               value = kwargs[col]
            self.instance_cols[col].set_value(value)
            
    def __repr__(self):
        return self.as_json()
    
    def __dict__(self):
        _dict = {}
        for col in self.get_columns():
            _dict[col] = self.instance_cols[col].get_value()
        return _dict
    
    def get_columns(self):
        return self.columns
    
    def get_values(self):
        return [self.instance_cols[col].get_value() for col in self.get_columns()]
    
    def get_table_name(self):
        return self.table_name

    def create_table(self):
        # Pypika's create table doesn't create valid MySQL for some reason so I'm doing a custom one
        column_defs = []
        for col in self.get_columns():
            column_defs.append(f'{self.instance_cols[col].get_def_statement(col)}')
        sql =  f'CREATE TABLE IF NOT EXISTS {self.get_table_name()} ({", ".join(column_defs)});'
        return self.__execute_with_wakeup(sql)

    def create(self):
        self.__validate_columns()
        table = Table(self.get_table_name())
        q = Query.into(table)
        q = self.__build_insert_query(q)
        response = self.__execute_with_wakeup(q.get_sql())
        response = json.dumps(response)
        pk = self.instance_cols[self.primary_key].get_value()
        if pk is None:
            key = self.instance_cols[self.primary_key].get_aws_value_type()
            pk = response['generatedFields'][0][key]
        return self.get_by_pk(pk)
    
    def save(self, where=None, fields=None):
        if where is None:
            where = self.primary_key
        if fields is None:
            fields = self.get_columns()
        self.__validate_columns()
        table = Table(self.get_table_name())
        q = Query.update(table)
        q = self.__build_update_query(q, fields)
        q = self.__build_in_where_clause(q, where, [self.instance_cols[where].get_value()])
        response = self.__execute_with_wakeup(q.get_sql())
        return self.get_by_pk()

    def as_json(self):
       return json.dumps(self.__dict__())

    def set_value(self, col, val):
        self.instance_cols[col].set_value(val)

    def get_by_pk(self, pks=None, fields=None, order_by=None, order="desc"):
        if pks is None:
            pks = self.instance_cols[self.primary_key].get_value()
        if pks is None:
            raise Exception('You need to pass value(s) or have one in the primary key of this object')
        return self.get_by_column(col=self.primary_key, vals=pks, fields=fields, order_by=order_by, order=order)
        
    def get_by_column(self, col=None, vals=None, fields=None, order_by=None, order="desc"):
        if fields is None:
            fields = self.get_columns()
        q = self.__build_from_table()
        q = self.__build_get_query(q, fields)
        q = self.__build_in_where_clause(q, col, vals)
        if order_by is not None:
            q = self.__build_order_by(q, order_by, order)
        response = self.__execute_with_wakeup(q.get_sql())
        response = json.dumps(response)
        records = self.__parse_records(response['records'], fields)
        return records

    def get_by_pypika_query(self, q):
        response = self.__execute_with_wakeup(q.get_sql(), with_meta=True)
        response = json.dumps(response)
        column_names = self.__parse_metadata(response['columnMetadata'])
        records = self.__parse_records(response['records'], column_names)
        return records

    def delete(self, col=None, vals=None):
        if col is None:
            col = self.primary_key
        if vals is None:
            vals = [self.instance_cols[self.primary_key].get_value()]
        q = self.__build_from_table()
        q = q.delete()
        q = self.__build_in_where_clause(q, col, vals)
        return self.__execute_with_wakeup(q.get_sql())

    def exec_raw_sql(self, sql):
        response = self.__execute_with_wakeup(sql, with_meta=True)
        return json.dumps(response)

    def __build_order_by(self, q, col=None, order="desc"):
        if col is None:
            col = self.primary_key
        o = Order[order]
        return q.orderby(col, o)

    def __build_from_table(self):
        table = Table(self.get_table_name())
        return Query.from_(table)

    def __parse_records(self, records=None, fields=None):
        parsed = []
        for record in records:
            parsed.append(self.__parse_record(record), fields)
        return parsed
        
    # Always return a new instance to avoid errors
    def __parse_record(self, record=None, fields=None):
        m = self.__class__()
        for i in range(0, len(record)):
            col = fields[i]
            if 'isNull' in record[i].keys():
                val = None
            else:
                val = record[i][self.instance_cols[col].get_aws_value_type()]
            m.set_value(col, val)
        return m
    
    def __parse_metadata(self, metadata):
        column_names = []
        for meta in metadata:
            if self.instance_cols[meta['name']] is not None:
                column_names.append(meta['name'])
        return column_names

    def __build_get_query(self, q, fields):
        return q.select(*fields)

    def __validate_columns(self):
        for col in self.get_columns():
            self.instance_cols[col].validate(col)

    def __build_insert_query(self, q):
        cols = tuple(self.get_columns())
        vals = tuple(self.get_values())
        return q.columns(*cols).insert(*vals)
    
    def __build_update_query(self, q, fields):
        table = Table(self.get_table_name())
        for col in fields:
            if col != self.primary_key:
                q = q.set(table[col],self.instance_cols[col].get_value())
        return q

    def __build_in_where_clause(self, q, col, vals):
        table = Table(self.get_table_name())
        vals = tuple(vals)
        return q.where(table[col].isin(vals))
    
    def __get_table_name(self):
        if('table_name' not in dir(self)):
            raise Exception('You must provide a table_name')
        return getattr(self, 'table_name')

    def __get_database_name(self):
        if('database' not in dir(self)):
            raise Exception('You must provide a database')
        return getattr(self, 'database')

    def __get_secret_arn(self):
        if('secret_arn' not in dir(self)):
            raise Exception('You must provide a secret_arn')
        return getattr(self, 'secret_arn')

    def __get_resource_arn(self):
        if('resource_arn' not in dir(self)):
            raise Exception('You must provide a resource_arn')
        return getattr(self, 'resource_arn')

    def __filter_columns(self):
        return [key for key in dir(self) if key[0] != '_' and isinstance(getattr(self, key), Col)]
    
    def __execute_with_wakeup(self, sql, with_meta=False):
        delay = 3
        max_attempts = 10
        attempt = 0

        while attempt < max_attempts:
            attempt += 1
            try:
                return self.__execute(sql, with_meta)
            except ClientError as ce:
                error_code = ce.response.get("Error").get('Code')
                error_msg = ce.response.get("Error").get('Message')

                # Aurora serverless is waking up
                if error_code == 'BadRequestException' and 'Communications link failure' in error_msg:
                    time.sleep(delay)
                else:
                    raise ce

        raise Exception('Waited for RDS Data but still getting error')
    
    def __execute(self, sql, with_meta=False):
        return client.execute_statement(
            continueAfterTimeout=True,
            database=self.database,
            resourceArn=self.resource_arn,
            secretArn=self.secret_arn,
            includeResultMetadata=with_meta,
            sql=sql
        )
        
        



from datetime import datetime
import json
class Col:
    def __init__(self, col_type, default_value=None, nullable=True, primary_key=False, unique=None, on_update=None):
        self.col_type = col_type
        self.default_value = default_value
        self.nullable = False if primary_key else nullable
        self.primary_key = primary_key
        self.value = self.set_value(default_value)
        self.unique = unique if unique is not None else self.primary_key
        self.on_update = on_update

    def get_col_type(self):
        return self.col_type

    def get_default_value(self):
        return self.default_value

    def is_nullable(self):
        return self.nullable

    def is_unique(self):
        return self.unique
    
    def is_primary_key(self):
        return self.primary_key

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def validate(self, key):
        if self.get_value() is None and not self.is_nullable() and self.default_value is None and self.is_auto() is not True:
            raise Exception(f'Can\'t set non nullable value to none for {str(key)}')
    
    def get_def_statement(self, name):
        val = f'{self.__get_initial_def(name)}'.strip()
        val = f'{val} {self.get_extras_for_def()}'.strip()
        return f'{val} {self.__get_contraints()}'.strip()

    def get_extras_for_def(self):
        return ''

    def is_auto(self):
        return False

    def __repr__(self):
        return str(self.value)

    def __get_initial_def(self, name):
        val = f'{name} {self.get_col_type()} {"" if self.is_nullable() else "NOT NULL"}'.strip()
        d = '' if self.default_value is None else f'DEFAULT {self.get_default_value()}'
        return f'{val} {d}'.strip()

    def __get_contraints(self):
        pk = 'PRIMARY KEY' if self.is_primary_key() else ''
        unique = 'UNIQUE' if self.is_unique() else ''
        on_update = f'ON UPDATE {self.on_update}' if self.on_update is not None else ''
        val = f'{pk} {unique}'.strip()
        return f'{val} {on_update}'.strip()

class NumericColumnType(Col):
    def __init__(self, col_type=None, auto_increment=False, default_value=None, nullable=True, primary_key=False, unique=None, unsigned=False, on_update=None):
        self.auto_increment = auto_increment
        self.unsigned = unsigned
        super().__init__(col_type=col_type, default_value=default_value, nullable=nullable, primary_key=primary_key, unique=unique, on_update=on_update)
    
    def get_extras_for_def(self):
        ai = 'AUTO_INCREMENT' if self.auto_increment else ''
        unsigned = 'UNSIGNED' if self.unsigned else ''
        return f'{ai} {unsigned}'.strip()
    
    def is_auto(self):
        return self.auto_increment

class Decimal(NumericColumnType):

    col_type = 'DECIMAL'

    def __init__(self, precision=10, digits=0, default_value=None, nullable=True, primary_key=False, unique=None, on_update=None):
        self.precision = precision
        self.digits = digits
        super().__init__(col_type=Decimal.col_type, nullable=nullable, primary_key=primary_key, unique=unique, default_value=default_value, on_update=on_update)
    
    def get_aws_value_type(self):
        return 'doubleValue'

    def get_col_type(self):
        return f'DECIMAL({self.precision},{self.digits})'.strip()

class IntColumn(NumericColumnType):

    col_type = 'INT'
    
    def __init__(self, default_value=None, nullable=True, primary_key=False, auto_increment=False, unique=None, on_update=None):
        super().__init__(col_type=IntColumn.col_type, auto_increment=auto_increment, default_value=default_value, nullable=nullable, primary_key=primary_key, unique=unique, on_update=on_update)
    
    def get_aws_value_type(self):
        return 'longValue'


class SmallInt(NumericColumnType):
    col_type = 'SMALLINT'

    def __init__(self, default_value=None, nullable=True, primary_key=False, auto_increment=False, unique=None, on_update=None):
        super().__init__(col_type=SmallInt.col_type, auto_increment=auto_increment, default_value=default_value, nullable=nullable, primary_key=primary_key, unique=unique, on_update=on_update)
    
    def get_aws_value_type(self):
        return 'longValue'

class Bool(Col):

    col_type = 'BOOLEAN'

    def __init__(self, default_value=None, nullable=True, on_update=None):
        super().__init__(col_type=Bool.col_type, default_value=default_value, nullable=nullable, on_update=on_update)

    def get_aws_value_type(self):
        return 'booleanValue'

    def set_value(self, val):
        if val is None:
            super().set_value(None)
        else:
            super().set_value(bool(val))

class VarChar(Col):

    col_type = 'VARCHAR'

    def __init__(self, size=100, default_value=None, nullable=True, primary_key=False, unique=None, on_update=None):
        self.size = size 
        super().__init__(col_type=VarChar.col_type, default_value=default_value, nullable=nullable, primary_key=primary_key, unique=unique, on_update=on_update)

    def get_size(self):
        return self.size

    def get_col_type(self):
        return f'VARCHAR({str(self.size)})'

    def get_aws_value_type(self):
        return 'stringValue'

class DateTime(Col):

    col_type = 'DATETIME(6)'

    def __init__(self, default_value=None, nullable=True, primary_key=False, unique=None, on_update=None):
        super().__init__(col_type=DateTime.col_type, default_value=default_value, nullable=nullable, primary_key=primary_key, unique=unique, on_update=on_update)
    
    # Always just get the max precision
    def get_col_type(self):
        return f'DATETIME(6)'

    def get_aws_value_type(self):
        return 'stringValue'
    
    def set_value(self, val):
        if val is None or 'CURRENT_TIMESTAMP' in val:
            super().set_value(None)
        else:
            super().set_value(datetime.strptime(val, "%Y-%m-%d %H:%M:%S.%f"))

class JsonArray(Col):

    col_type = 'JSON'

    def __init__(self, unique=None, on_update=None):
        super().__init__(col_type=JsonArray.col_type, default_value=None, nullable=True, unique=unique, on_update=on_update)

    def get_col_type(self):
        return 'JSON'

    def get_aws_value_type(self):
        return 'stringValue'

class JsonObject(Col):

    col_type = 'JSON'

    def __init__(self, unique=None, on_update=None):
        super().__init__(col_type=JsonObject.col_type, default_value=None, nullable=True, unique=unique, on_update=on_update)

    def get_col_type(self):
        return 'JSON'

    def get_aws_value_type(self):
        return 'stringValue'
    
    def get_default_value(self):
        return f'JSON_OBJECT{tuple(self.default_value)}'
    

